/*
 * Copyright 1993-2009 NVIDIA Corporation.  All rights reserved.
 *
 * NVIDIA Corporation and its licensors retain all intellectual property and 
 * proprietary rights in and to this software and related documentation and 
 * any modifications thereto.  Any use, reproduction, disclosure, or distribution 
 * of this software and related documentation without an express license 
 * agreement from NVIDIA Corporation is strictly prohibited.
 * 
 */



#include <assert.h>
#include <cutil_inline.h>
#include "sortingNetworks_common.cuh"
#include "sortingNetworks_common.h"



////////////////////////////////////////////////////////////////////////////////
// Monolithic bitonic sort kernel for short arrays fitting into shared memory
////////////////////////////////////////////////////////////////////////////////
__global__ void bitonicSortShared(
    unsigned int *d_DstKey,
    unsigned int *d_DstVal,
    unsigned int *d_SrcKey,
    unsigned int *d_SrcVal,
    unsigned int arrayLength,
    unsigned int dir
){
    //Shared memory storage for one or more short vectors
    __shared__ unsigned int s_key[SHARED_SIZE_LIMIT];
    __shared__ unsigned int s_val[SHARED_SIZE_LIMIT];

    //Offset to the beginning of subbatch and load data
    d_SrcKey += blockIdx.x * SHARED_SIZE_LIMIT + threadIdx.x;
    d_SrcVal += blockIdx.x * SHARED_SIZE_LIMIT + threadIdx.x;
    d_DstKey += blockIdx.x * SHARED_SIZE_LIMIT + threadIdx.x;
    d_DstVal += blockIdx.x * SHARED_SIZE_LIMIT + threadIdx.x;
    s_key[threadIdx.x +                       0] = d_SrcKey[                      0];
    s_val[threadIdx.x +                       0] = d_SrcVal[                      0];
    s_key[threadIdx.x + (SHARED_SIZE_LIMIT / 2)] = d_SrcKey[(SHARED_SIZE_LIMIT / 2)];
    s_val[threadIdx.x + (SHARED_SIZE_LIMIT / 2)] = d_SrcVal[(SHARED_SIZE_LIMIT / 2)];

    unsigned int size = 2;
    for(; size < arrayLength; size <<= 1){
        //Bitonic merge
        unsigned int ddd = dir ^ ( (threadIdx.x & (size / 2)) != 0 );
        for(unsigned int stride = size / 2; stride > 0; stride >>= 1){
            __syncthreads();
            unsigned int pos = 2 * threadIdx.x - (threadIdx.x & (stride - 1));
            Comparator(
                s_key[pos +      0], s_val[pos +      0],
                s_key[pos + stride], s_val[pos + stride],
                ddd
            );
        }
    }

    //ddd == dir for last merge step
    for(unsigned int stride = size / 2; stride > 0; stride >>= 1){
        __syncthreads();
        unsigned int pos = 2 * threadIdx.x - (threadIdx.x & (stride - 1));
        Comparator(
            s_key[pos +      0], s_val[pos +      0],
            s_key[pos + stride], s_val[pos + stride],
            dir
        );
    }

    __syncthreads();
    d_DstKey[                      0] = s_key[threadIdx.x +                       0];
    d_DstVal[                      0] = s_val[threadIdx.x +                       0];
    d_DstKey[(SHARED_SIZE_LIMIT / 2)] = s_key[threadIdx.x + (SHARED_SIZE_LIMIT / 2)];
    d_DstVal[(SHARED_SIZE_LIMIT / 2)] = s_val[threadIdx.x + (SHARED_SIZE_LIMIT / 2)];
}



////////////////////////////////////////////////////////////////////////////////
// Bitonic sort kernel for large arrays (not fitting into shared memory)
////////////////////////////////////////////////////////////////////////////////
//Combined bitonic merge step for 'size' = [2 .. SHARED_SIZE_LIMIT]
//Could actually use bitonicSortShared for this step
//but need even / odd subarrays to be sorted in oppsite directions
__global__ void bitonicMergeShared1(
    unsigned int *d_DstKey,
    unsigned int *d_DstVal,
    unsigned int *d_SrcKey,
    unsigned int *d_SrcVal,
    unsigned int arrayLength,
    unsigned int dir
){
    //Shared memory storage for current subarray
    __shared__ unsigned int s_key[SHARED_SIZE_LIMIT];
    __shared__ unsigned int s_val[SHARED_SIZE_LIMIT];

    //Offset to the beginning of subarray and load data
    d_SrcKey += blockIdx.x * SHARED_SIZE_LIMIT + threadIdx.x;
    d_SrcVal += blockIdx.x * SHARED_SIZE_LIMIT + threadIdx.x;
    d_DstKey += blockIdx.x * SHARED_SIZE_LIMIT + threadIdx.x;
    d_DstVal += blockIdx.x * SHARED_SIZE_LIMIT + threadIdx.x;
    s_key[threadIdx.x +                       0] = d_SrcKey[                      0];
    s_val[threadIdx.x +                       0] = d_SrcVal[                      0];
    s_key[threadIdx.x + (SHARED_SIZE_LIMIT / 2)] = d_SrcKey[(SHARED_SIZE_LIMIT / 2)];
    s_val[threadIdx.x + (SHARED_SIZE_LIMIT / 2)] = d_SrcVal[(SHARED_SIZE_LIMIT / 2)];

    for(unsigned int size = 2; size < SHARED_SIZE_LIMIT; size <<= 1){
        //Bitonic merge
        unsigned int ddd = dir ^ ( (threadIdx.x & (size / 2)) != 0 );
        for(unsigned int stride = size / 2; stride > 0; stride >>= 1){
            __syncthreads();
            unsigned int pos = 2 * threadIdx.x - (threadIdx.x & (stride - 1));
            Comparator(
                s_key[pos +      0], s_val[pos +      0],
                s_key[pos + stride], s_val[pos + stride],
                ddd
            );
        }
    }

    {
        //Bitonic merge
        unsigned int ddd = dir ^ (blockIdx.x & 1);
        for(unsigned int stride = SHARED_SIZE_LIMIT / 2; stride > 0; stride >>= 1){
            __syncthreads();
            unsigned int pos = 2 * threadIdx.x - (threadIdx.x & (stride - 1));
            Comparator(
                s_key[pos +      0], s_val[pos +      0],
                s_key[pos + stride], s_val[pos + stride],
                ddd
            );
        }
    }


    __syncthreads();
    d_DstKey[                      0] = s_key[threadIdx.x +                       0];
    d_DstVal[                      0] = s_val[threadIdx.x +                       0];
    d_DstKey[(SHARED_SIZE_LIMIT / 2)] = s_key[threadIdx.x + (SHARED_SIZE_LIMIT / 2)];
    d_DstVal[(SHARED_SIZE_LIMIT / 2)] = s_val[threadIdx.x + (SHARED_SIZE_LIMIT / 2)];
}

//Bitonic merge step for stride < SHARED_SIZE_LIMIT
__global__ void bitonicMergeShared2(
    unsigned int *d_DstKey,
    unsigned int *d_DstVal,
    unsigned int *d_SrcKey,
    unsigned int *d_SrcVal,
    unsigned int arrayLength,
    unsigned int size,
    unsigned int dir
){
    //Shared memory storage for current subarray
    __shared__ unsigned int s_key[SHARED_SIZE_LIMIT];
    __shared__ unsigned int s_val[SHARED_SIZE_LIMIT];

    d_SrcKey += blockIdx.x * SHARED_SIZE_LIMIT + threadIdx.x;
    d_SrcVal += blockIdx.x * SHARED_SIZE_LIMIT + threadIdx.x;
    d_DstKey += blockIdx.x * SHARED_SIZE_LIMIT + threadIdx.x;
    d_DstVal += blockIdx.x * SHARED_SIZE_LIMIT + threadIdx.x;
    s_key[threadIdx.x +                       0] = d_SrcKey[                      0];
    s_val[threadIdx.x +                       0] = d_SrcVal[                      0];
    s_key[threadIdx.x + (SHARED_SIZE_LIMIT / 2)] = d_SrcKey[(SHARED_SIZE_LIMIT / 2)];
    s_val[threadIdx.x + (SHARED_SIZE_LIMIT / 2)] = d_SrcVal[(SHARED_SIZE_LIMIT / 2)];

    //Bitonic merge
    unsigned int comparatorI = UMAD(blockIdx.x, blockDim.x, threadIdx.x) & ((arrayLength / 2) - 1);
    unsigned int ddd = dir ^ ( (comparatorI & (size / 2)) != 0 );
    for(unsigned int stride = SHARED_SIZE_LIMIT / 2; stride > 0; stride >>= 1){
        __syncthreads();
        unsigned int pos = 2 * threadIdx.x - (threadIdx.x & (stride - 1));
        Comparator(
            s_key[pos +      0], s_val[pos +      0],
            s_key[pos + stride], s_val[pos + stride],
            ddd
        );
    }

    __syncthreads();
    d_DstKey[                      0] = s_key[threadIdx.x +                       0];
    d_DstVal[                      0] = s_val[threadIdx.x +                       0];
    d_DstKey[(SHARED_SIZE_LIMIT / 2)] = s_key[threadIdx.x + (SHARED_SIZE_LIMIT / 2)];
    d_DstVal[(SHARED_SIZE_LIMIT / 2)] = s_val[threadIdx.x + (SHARED_SIZE_LIMIT / 2)];
}

//Bitonic merge iteration for stride >= SHARED_SIZE_LIMIT
__global__ void bitonicMergeGlobal(
    unsigned int *d_DstKey,
    unsigned int *d_DstVal,
    unsigned int *d_SrcKey,
    unsigned int *d_SrcVal,
    unsigned int arrayLength,
    unsigned int size,
    unsigned int stride,
    unsigned int dir
){
    unsigned int global_comparatorI = blockIdx.x * blockDim.x + threadIdx.x;
    unsigned int        comparatorI = global_comparatorI & (arrayLength / 2 - 1);

    //Bitonic merge
    unsigned int ddd = dir ^ ( (comparatorI & (size / 2)) != 0 );
    unsigned int pos = 2 * global_comparatorI - (global_comparatorI & (stride - 1));

    unsigned int keyA = d_SrcKey[pos +      0];
    unsigned int valA = d_SrcVal[pos +      0];
    unsigned int keyB = d_SrcKey[pos + stride];
    unsigned int valB = d_SrcVal[pos + stride];

    Comparator(
        keyA, valA,
        keyB, valB,
        ddd
    );

    d_DstKey[pos +      0] = keyA;
    d_DstVal[pos +      0] = valA;
    d_DstKey[pos + stride] = keyB;
    d_DstVal[pos + stride] = valB;
}



////////////////////////////////////////////////////////////////////////////////
// Interface function
////////////////////////////////////////////////////////////////////////////////
extern "C" void bitonicSort(
    unsigned int *d_DstKey,
    unsigned int *d_DstVal,
    unsigned int *d_SrcKey,
    unsigned int *d_SrcVal,
    unsigned int batchSize,
    unsigned int arrayLength,
    unsigned int dir
){
    unsigned int  blockCount = batchSize * arrayLength / SHARED_SIZE_LIMIT;
    unsigned int threadCount = SHARED_SIZE_LIMIT / 2;

    if(arrayLength <= SHARED_SIZE_LIMIT){
        assert( SHARED_SIZE_LIMIT % arrayLength == 0 );
        bitonicSortShared<<<blockCount, threadCount>>>(d_DstKey, d_DstVal, d_SrcKey, d_SrcVal, arrayLength, dir);
    }else{
        bitonicMergeShared1<<<blockCount, threadCount>>>(d_DstKey, d_DstVal, d_SrcKey, d_SrcVal, arrayLength, dir);

        for(unsigned int size = 2 * SHARED_SIZE_LIMIT; size <= arrayLength; size <<= 1)
            for(unsigned stride = size / 2; stride > 0; stride >>= 1)
                if(stride >= SHARED_SIZE_LIMIT){
                    bitonicMergeGlobal<<<batchSize * arrayLength / 512, 256>>>(d_DstKey, d_DstVal, d_DstKey, d_DstVal, arrayLength, size, stride, dir);
                }else{
                    bitonicMergeShared2<<<blockCount, threadCount>>>(d_DstKey, d_DstVal, d_DstKey, d_DstVal, arrayLength, size, dir);
                    break;
                }
    }
}


extern "C" int isDeviceEmulation(void){
    #ifdef __DEVICE_EMULATION__
        return 1;
    #else
        return 0;
    #endif
}
