/*
 * Copyright 1993-2009 NVIDIA Corporation.  All rights reserved.
 *
 * NVIDIA Corporation and its licensors retain all intellectual property and 
 * proprietary rights in and to this software and related documentation and 
 * any modifications thereto.  Any use, reproduction, disclosure, or distribution 
 * of this software and related documentation without an express license 
 * agreement from NVIDIA Corporation is strictly prohibited.
 * 
 */



#include <assert.h>
#include <cutil_inline.h>
#include "sortingNetworks_common.cuh"
#include "sortingNetworks_common.h"



////////////////////////////////////////////////////////////////////////////////
// Monolithic Bacther's sort kernel for short arrays fitting into shared memory
////////////////////////////////////////////////////////////////////////////////
__global__ void oddEvenMergeSortShared(
    unsigned int *d_DstKey,
    unsigned int *d_DstVal,
    unsigned int *d_SrcKey,
    unsigned int *d_SrcVal,
    unsigned int arrayLength,
    unsigned int dir
){
    //Shared memory storage for one or more small vectors
    __shared__ unsigned int s_key[SHARED_SIZE_LIMIT];
    __shared__ unsigned int s_val[SHARED_SIZE_LIMIT];

    //Offset to the beginning of subbatch and load data
    d_SrcKey += blockIdx.x * SHARED_SIZE_LIMIT + threadIdx.x;
    d_SrcVal += blockIdx.x * SHARED_SIZE_LIMIT + threadIdx.x;
    d_DstKey += blockIdx.x * SHARED_SIZE_LIMIT + threadIdx.x;
    d_DstVal += blockIdx.x * SHARED_SIZE_LIMIT + threadIdx.x;
    s_key[threadIdx.x +                       0] = d_SrcKey[                      0];
    s_val[threadIdx.x +                       0] = d_SrcVal[                      0];
    s_key[threadIdx.x + (SHARED_SIZE_LIMIT / 2)] = d_SrcKey[(SHARED_SIZE_LIMIT / 2)];
    s_val[threadIdx.x + (SHARED_SIZE_LIMIT / 2)] = d_SrcVal[(SHARED_SIZE_LIMIT / 2)];

    for(unsigned int size = 2; size <= arrayLength; size <<= 1){
        unsigned int stride = size / 2;
        unsigned int offset = threadIdx.x & (stride - 1);

        {
            __syncthreads();
            unsigned int pos = 2 * threadIdx.x - (threadIdx.x & (stride - 1));
            Comparator(
                s_key[pos +      0], s_val[pos +      0],
                s_key[pos + stride], s_val[pos + stride],
                dir
            );
            stride >>= 1;
        }

        for(; stride > 0; stride >>= 1){
            __syncthreads();
            unsigned int pos = 2 * threadIdx.x - (threadIdx.x & (stride - 1));
            if(offset >= stride)
                Comparator(
                    s_key[pos - stride], s_val[pos - stride],
                    s_key[pos +      0], s_val[pos +      0],
                    dir
                );
        }
    }

    __syncthreads();
    d_DstKey[                      0] = s_key[threadIdx.x +                       0];
    d_DstVal[                      0] = s_val[threadIdx.x +                       0];
    d_DstKey[(SHARED_SIZE_LIMIT / 2)] = s_key[threadIdx.x + (SHARED_SIZE_LIMIT / 2)];
    d_DstVal[(SHARED_SIZE_LIMIT / 2)] = s_val[threadIdx.x + (SHARED_SIZE_LIMIT / 2)];
}



////////////////////////////////////////////////////////////////////////////////
// Batcher's sort kernel for large arrays (not fitting into shared memory)
////////////////////////////////////////////////////////////////////////////////
//Bitonic merge step for stride < SHARED_SIZE_LIMIT
__global__ void oddEvenMergeShared2(
    unsigned int *d_DstKey,
    unsigned int *d_DstVal,
    unsigned int *d_SrcKey,
    unsigned int *d_SrcVal,
    unsigned int arrayLength,
    unsigned int size,
    unsigned int dir
){
    //Shared memory storage for current subarray
    __shared__ unsigned int s_key[SHARED_SIZE_LIMIT];
    __shared__ unsigned int s_val[SHARED_SIZE_LIMIT];

    d_SrcKey += blockIdx.x * SHARED_SIZE_LIMIT + threadIdx.x;
    d_SrcVal += blockIdx.x * SHARED_SIZE_LIMIT + threadIdx.x;
    d_DstKey += blockIdx.x * SHARED_SIZE_LIMIT + threadIdx.x;
    d_DstVal += blockIdx.x * SHARED_SIZE_LIMIT + threadIdx.x;
    s_key[threadIdx.x +                       0] = d_SrcKey[                      0];
    s_val[threadIdx.x +                       0] = d_SrcVal[                      0];
    s_key[threadIdx.x + (SHARED_SIZE_LIMIT / 2)] = d_SrcKey[(SHARED_SIZE_LIMIT / 2)];
    s_val[threadIdx.x + (SHARED_SIZE_LIMIT / 2)] = d_SrcVal[(SHARED_SIZE_LIMIT / 2)];

    __syncthreads();
    d_DstKey[                      0] = s_key[threadIdx.x +                       0];
    d_DstVal[                      0] = s_val[threadIdx.x +                       0];
    d_DstKey[(SHARED_SIZE_LIMIT / 2)] = s_key[threadIdx.x + (SHARED_SIZE_LIMIT / 2)];
    d_DstVal[(SHARED_SIZE_LIMIT / 2)] = s_val[threadIdx.x + (SHARED_SIZE_LIMIT / 2)];
}



//Odd-even merge iteration for stride >= SHARED_SIZE_LIMIT
__global__ void oddEvenMergeGlobal(
    unsigned int *d_DstKey,
    unsigned int *d_DstVal,
    unsigned int *d_SrcKey,
    unsigned int *d_SrcVal,
    unsigned int arrayLength,
    unsigned int size,
    unsigned int stride,
    unsigned int dir
){
    unsigned int global_comparatorI = blockIdx.x * blockDim.x + threadIdx.x;

    //Odd-even merge
    unsigned int pos = 2 * global_comparatorI - (global_comparatorI & (stride - 1));

    if(stride < size / 2){
        unsigned int offset = global_comparatorI & ( (size / 2) - 1);
        if(offset >= stride){
            unsigned int keyA = d_SrcKey[pos - stride];
            unsigned int valA = d_SrcVal[pos - stride];
            unsigned int keyB = d_SrcKey[pos +      0];
            unsigned int valB = d_SrcVal[pos +      0];

            Comparator(
                keyA, valA,
                keyB, valB,
                dir
            );

            d_DstKey[pos - stride] = keyA;
            d_DstVal[pos - stride] = valA;
            d_DstKey[pos +      0] = keyB;
            d_DstVal[pos +      0] = valB;
        }
    }else{
        unsigned int keyA = d_SrcKey[pos +      0];
        unsigned int valA = d_SrcVal[pos +      0];
        unsigned int keyB = d_SrcKey[pos + stride];
        unsigned int valB = d_SrcVal[pos + stride];

        Comparator(
            keyA, valA,
            keyB, valB,
            dir
        );

        d_DstKey[pos +      0] = keyA;
        d_DstVal[pos +      0] = valA;
        d_DstKey[pos + stride] = keyB;
        d_DstVal[pos + stride] = valB;
    }
}



////////////////////////////////////////////////////////////////////////////////
// Interface function
////////////////////////////////////////////////////////////////////////////////
extern "C" void oddEvenMergeSort(
    unsigned int *d_DstKey,
    unsigned int *d_DstVal,
    unsigned int *d_SrcKey,
    unsigned int *d_SrcVal,
    unsigned int batchSize,
    unsigned int arrayLength,
    unsigned int dir
){
    unsigned int  blockCount = batchSize * arrayLength / SHARED_SIZE_LIMIT;
    unsigned int threadCount = SHARED_SIZE_LIMIT / 2;

    if(arrayLength <= SHARED_SIZE_LIMIT){
        assert( SHARED_SIZE_LIMIT % arrayLength == 0 );
        oddEvenMergeSortShared<<<blockCount, threadCount>>>(d_DstKey, d_DstVal, d_SrcKey, d_SrcVal, arrayLength, dir);
    }else{
        oddEvenMergeSortShared<<<blockCount, threadCount>>>(d_DstKey, d_DstVal, d_SrcKey, d_SrcVal, SHARED_SIZE_LIMIT, dir);

        for(unsigned int size = 2 * SHARED_SIZE_LIMIT; size <= arrayLength; size <<= 1)
            for(unsigned stride = size / 2; stride > 0; stride >>= 1)
                if(stride >= 1){
                    oddEvenMergeGlobal<<<batchSize * arrayLength / 512, 256>>>(d_DstKey, d_DstVal, d_DstKey, d_DstVal, arrayLength, size, stride, dir);
                }else{
                    oddEvenMergeShared2<<<blockCount, threadCount>>>(d_DstKey, d_DstVal, d_DstKey, d_DstVal, arrayLength, size, dir);
                    break;
                }
    }
}
