/*
 * Copyright 1993-2009 NVIDIA Corporation.  All rights reserved.
 *
 * NVIDIA Corporation and its licensors retain all intellectual property and 
 * proprietary rights in and to this software and related documentation and 
 * any modifications thereto.  Any use, reproduction, disclosure, or distribution 
 * of this software and related documentation without an express license 
 * agreement from NVIDIA Corporation is strictly prohibited.
 * 
 */
 

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "sortingNetworks_common.h"



////////////////////////////////////////////////////////////////////////////////
// Validate sorted keys array (check for integrity and proper order)
////////////////////////////////////////////////////////////////////////////////
extern "C" unsigned int validateSortedKeys(
    unsigned int *resKey,
    unsigned int *srcKey,
    unsigned int batchSize,
    unsigned int arrayLength,
    unsigned int numValues,
    unsigned int dir
){
    unsigned int *srcHist;
    unsigned int *resHist;
    if(arrayLength < 2){
        printf("validateSortedKeys(): arrayLength too short, exiting...\n");
        return 1;
    }

    srcHist = (unsigned int *)malloc(numValues * sizeof(unsigned int));
    resHist = (unsigned int *)malloc(numValues * sizeof(unsigned int));

    int flag = 1;
    for(unsigned int j = 0; j < batchSize; j++, srcKey += arrayLength, resKey += arrayLength){
        //Build histograms for keys arrays
        memset(srcHist, 0, numValues * sizeof(unsigned int));
        memset(resHist, 0, numValues * sizeof(unsigned int));
        for(unsigned int i = 0; i < arrayLength; i++){
            if( srcKey[i] < numValues && resKey[i] < numValues ){
                srcHist[srcKey[i]]++;
                resHist[resKey[i]]++;
            }else{
                flag = 0;
                break;
            }
        }

        if(!flag){
            printf("validateSortedKeys(): ***Set %u source/result key arrays are not limited properly***\n", j);
            goto brk;
        }

        //Compare the histograms
        for(unsigned int i = 0; i < numValues; i++)
            if(srcHist[i] != resHist[i]){
                flag = 0;
                break;
            }

        if(!flag){
            printf("validateSortedKeys(): ***Set %u source/result keys histograms do not match***\n", j);
            goto brk;
        }

        if(dir){
            //Ascending order
            for(unsigned int i = 0; i < arrayLength - 1; i++)
                if(resKey[i + 1] < resKey[i]){
                    flag = 0;
                    break;
                }
        }else{
            //Descending order
            for(unsigned int i = 0; i < arrayLength - 1; i++)
                if(resKey[i + 1] > resKey[i]){
                    flag = 0;
                    break;
                }
        }

        if(!flag){
            printf("validateSortedKeys(): ***Set %u result key array is not ordered properly***\n", j);
            goto brk;
        }
    }

brk:
    free(resHist);
    free(srcHist);

    if(flag) printf("validateSortedKeys(): result key array is ordered correctly\n");
    return flag;
}



////////////////////////////////////////////////////////////////////////////////
// Generate input values from input keys
// Check output values to match output keys by the same translation function
////////////////////////////////////////////////////////////////////////////////
static unsigned int translateKey(unsigned int r){
    const unsigned int log2N = 32;

    r = ((r >>  1) & 0x55555555) | ((r & 0x55555555) <<  1);
    r = ((r >>  2) & 0x33333333) | ((r & 0x33333333) <<  2);
    r = ((r >>  4) & 0x0F0F0F0F) | ((r & 0x0F0F0F0F) <<  4);
    r = ((r >>  8) & 0x00FF00FF) | ((r & 0x00FF00FF) <<  8);
    r = ((r >> 16) & 0x0000FFFF) | ((r & 0x0000FFFF) << 16);
    return r >> (32 - log2N);
}

extern "C" void generateValuesFromKeys(
    unsigned int *ival,
    unsigned int *ikey,
    unsigned int N
){
    for(unsigned int i = 0; i < N; i++)
        ival[i] = translateKey(ikey[i]);
}

extern "C" int validateValues(
    unsigned int *oval,
    unsigned int *okey,
    unsigned int N
){
    int flag = 1;
    for(unsigned int i = 0; i < N; i++)
        if( oval[i] != translateKey(okey[i]) ){
            printf("validateValues(): ***val array doesn't match key array***\n");
            flag = 0;
            break;
        }

    if(flag) printf("validateValues(): key and val arrays match\n");
    return flag;
}








static void Comparator(unsigned int& a, unsigned int& b, unsigned int dir){
    if( (a > b) == dir ){
        unsigned int t = a;
        a = b; b = t;
    }
}

static void oddEvenMergeRecursion(unsigned int *data, unsigned int n, unsigned int stride, unsigned int dir){
    if(n > 2 * stride){
        oddEvenMergeRecursion(data,          n, 2 * stride, dir); //Even subsequence
        oddEvenMergeRecursion(data + stride, n, 2 * stride, dir); //Odd subsequence

        for(unsigned int i = stride; i < n - stride; i += 2 * stride)
            Comparator(data[i], data[i + stride], dir);
    }else
        Comparator(data[0], data[stride], dir);
}

extern "C" void oddEvenMergeSortRecursion(unsigned int *data, unsigned int n, unsigned int dir){
    if(n > 1){
        oddEvenMergeSortRecursion(data,         n / 2, dir);
        oddEvenMergeSortRecursion(data + n / 2, n / 2, dir);
        oddEvenMergeRecursion(data, n, 1, dir);
    }
}



static void oddEvenMergeHost(unsigned int *data, unsigned int size, unsigned int dir){
    //stride = size / 2
    for(unsigned int tid = 0; tid < size / 2; tid++)
        Comparator(data[tid], data[tid + size / 2], dir);

    for(unsigned int stride = size / 4; stride > 0; stride >>= 1){
        for(unsigned int tid = 0; tid < size / 2 - stride; tid++){
            unsigned int pos = 2 * tid - (tid & (stride - 1));
            Comparator(data[pos + stride], data[pos + 2 * stride], dir);
        }
    }
}

extern "C" void oddEvenMergeSortHost(unsigned int *data, unsigned int n, unsigned int dir){
    for(unsigned int size = 2; size <= n; size <<= 1){

        unsigned int stride = size / 2;
        {
            for(unsigned int comparatorI = 0; comparatorI < n / 2; comparatorI++){
                unsigned int pos = 2 * comparatorI - (comparatorI & (stride - 1));
                Comparator(data[pos], data[pos + stride], dir);
            }
        }

        stride >>= 1;
        for(; stride > 0; stride >>= 1){
            for(unsigned int comparatorI = 0; comparatorI < n / 2; comparatorI++){
                unsigned int   i = comparatorI & ( (size / 2) - 1);
                unsigned int pos = 2 * comparatorI - (comparatorI & (stride - 1));

                if(i >= stride)
                    Comparator(data[pos - stride], data[pos], dir);
            }
        }
    }
}


